import flax.linen as nn
import jax
import jax.lax as lax
import jax.numpy as jnp
from einops import rearrange
from functools import partial
from .flash_attention_tpu import _flash_attention_fwd, _flash_attention_bwd, BlockSizes
from .flash_attention_gpu import _mha_forward, _mha_backward


def _ring_attention_fwd(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    numerator = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(q.dtype)
    denominator = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
    axis_size = lax.psum(1, axis_name)
    block_size = q_len  # assumes this function is pre-sharded inside shard_map
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    def scan_kv_block(carry, idx):
        prev_max_score, numerator, denominator, k, v = carry
        ''' einsum version
        mask = lax.dynamic_slice_in_dim(attn_mask,
            (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        attn_weights = jnp.einsum("bqhd,bkd->bhqk", q, k) / scale
        attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        max_score = jnp.maximum(prev_max_score, jnp.max(attn_weights, axis=-1))
        exp_weights = jnp.exp(attn_weights - max_score[..., None])
        correction = rearrange(jnp.exp(prev_max_score - max_score), 'b h q -> b q h')[..., None]
        numerator = numerator * correction + jnp.einsum("bhqk,bkd->bqhd", exp_weights, v)
        denominator = denominator * jnp.exp(prev_max_score - max_score) + jnp.sum(exp_weights, axis=-1)
        '''
        attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
                                                   (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len,
                                                   axis=-1
                                                   )
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        numerator, denominator, max_score = _blockwise_attention_fwd(q, k, v, (numerator, denominator, prev_max_score),
                                                                     q_chunk_idx_start, k_chunk_idx_start,
                                                                     bias=attn_bias_slice, **blockwise_kwargs)
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]),
                   (k, v))
        return (max_score, numerator, denominator, k, v), None

    prev_max_score = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)
    (max_score, numerator, denominator, _, _), _ = lax.scan(scan_kv_block,
                                                            init=(prev_max_score, numerator, denominator, k, v),
                                                            xs=jnp.arange(0, axis_size))
    output = numerator / rearrange(denominator, 'b h q -> b q h')[..., None]
    return output.astype(v.dtype), (output, q, k, v, attn_bias, denominator, max_score)


def _ring_attention_bwd(axis_name, float32_logits, blockwise_kwargs, res, g):
    del float32_logits
    output, q, k, v, attn_bias, denominator, max_score = res
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    axis_size = lax.psum(1, axis_name)
    dq = jnp.zeros_like(q, dtype=q.dtype)
    dk = jnp.zeros_like(k, dtype=k.dtype)
    dv = jnp.zeros_like(v, dtype=k.dtype)
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]
    block_size = q.shape[1]  # assumes this function is pre-sharded inside shard_map

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        ''' einsum version
        # mask = lax.dynamic_slice_in_dim(attn_mask,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        # attn_weights = jnp.einsum("bqhd,bkd->bhqk", q, k) / scale
        # attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        # exp_weights = jnp.exp(attn_weights - max_score[..., None]) / denominator[..., None]
        # ds = jnp.einsum("bqhd,bkd->bhqk", g, v)
        # dl = (ds - jnp.einsum("bqhd,bqhd->bhs", g, output)[..., None]) * exp_weights
        # dq = dq + jnp.einsum("bhqk,bkd->bqhd", dl, k) / scale
        # dk = dk + jnp.einsum("bqhd,bhqk->bkd", q, dl) / scale
        # dv = dv + jnp.einsum("bhqk,bqhd->bkd", exp_weights, g)
        '''
        attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
                                                   (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len,
                                                   axis=-1
                                                   )
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        dq, dk, dv = _blockwise_attention_bwd(q, k, v, g, (dq, dk, dv, output, denominator, max_score),
                                              q_chunk_idx_start, k_chunk_idx_start, bias=attn_bias_slice,
                                              **blockwise_kwargs)
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(k.dtype)
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6])
def ring_attention(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    y, _ = _ring_attention_fwd(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs)
    return y


ring_attention.defvjp(_ring_attention_fwd, _ring_attention_bwd)


def _ring_attention_standard_fwd(q, k, v, attn_mask, axis_name, float32_logits):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    batch, q_len, num_heads, _ = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    numerator = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(q.dtype)
    denominator = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
    axis_size = lax.psum(1, axis_name)
    scale = jnp.sqrt(q.shape[-1])

    def scan_kv_block(carry, idx):
        prev_max_score, numerator, denominator, k, v = carry
        mask = lax.dynamic_slice_in_dim(attn_mask,
                                        (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        attn_weights = jnp.einsum("bqhd,bkhd->bhqk", q, k) / scale
        # attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        attn_weights = jnp.where(mask, attn_weights, -jnp.inf)
        max_score = jnp.maximum(prev_max_score, jnp.max(attn_weights, axis=-1))
        exp_weights = jnp.exp(attn_weights - max_score[..., None])
        correction = rearrange(jnp.exp(prev_max_score - max_score), 'b h q -> b q h')[..., None]
        numerator = numerator * correction + jnp.einsum("bhqk,bkhd->bqhd", exp_weights, v)
        denominator = denominator * jnp.exp(prev_max_score - max_score) + jnp.sum(exp_weights, axis=-1)
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                               (i + 1) % axis_size) for i in range(axis_size)]), (k, v))
        return (max_score, numerator, denominator, k, v), None

    prev_max_score = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)
    (max_score, numerator, denominator, _, _), _ = lax.scan(scan_kv_block,
                                                            init=(prev_max_score, numerator, denominator, k, v),
                                                            xs=jnp.arange(0, axis_size))
    output = numerator / rearrange(denominator, 'b h q -> b q h')[..., None]
    return output.astype(v.dtype), (output, q, k, v, attn_mask, numerator, denominator, max_score)


def _ring_attention_standard_bwd(axis_name, float32_logits, res, g):
    del float32_logits
    axis_size = lax.psum(1, axis_name)
    output, q, k, v, attn_mask, numerator, denominator, max_score = res
    dq = jnp.zeros_like(q, dtype=jnp.float32)
    dk = jnp.zeros_like(k, dtype=jnp.float32)
    dv = jnp.zeros_like(v, dtype=jnp.float32)
    batch, kv_len, num_heads, dim_per_head = k.shape
    scale = jnp.sqrt(q.shape[-1])

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        mask = lax.dynamic_slice_in_dim(attn_mask,
                                        (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        attn_weights = jnp.einsum("bqhd,bkhd->bhqk", q, k) / scale
        # attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        attn_weights = jnp.where(mask, attn_weights, -jnp.inf)
        exp_weights = jnp.exp(attn_weights - max_score[..., None]) / denominator[..., None]
        ds = jnp.einsum("bqhd,bkhd->bhqk", g, v)
        dl = (ds - jnp.einsum("bqhd,bqhd->bhq", g, output)[..., None]) * exp_weights
        dq = dq + jnp.einsum("bhqk,bkhd->bqhd", dl, k) / scale
        dk = dk + jnp.einsum("bqhd,bhqk->bkhd", q, dl) / scale
        dv = dv + jnp.einsum("bhqk,bqhd->bkhd", exp_weights, g)
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5])
def ring_attention_standard(q, k, v, attn_mask, axis_name, float32_logits=True):
    y, _ = _ring_attention_standard_fwd(q, k, v, attn_mask, axis_name, float32_logits)
    return y


ring_attention_standard.defvjp(_ring_attention_standard_fwd, _ring_attention_standard_bwd)


def _blockwise_attention_fwd(q, k, v, carry, q_chunk_idx_start, k_chunk_idx_start, bias, causal, query_chunk_size,
                             key_chunk_size, deterministic, dropout_rng, attn_pdrop, dtype, policy, precision,
                             prevent_cse):
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    batch, kv_len, num_heads, dim_per_head = v.shape
    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    k = k.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    v = v.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    q, k, v = map(lambda x: jnp.moveaxis(x, 1, 0), (q, k, v))

    numerator, denominator, max_score = carry
    numerator = numerator.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    numerator = jnp.moveaxis(numerator, 1, 0)
    denominator = denominator.reshape((batch, num_heads, num_q, query_chunk_size))
    max_score = max_score.reshape((batch, num_heads, num_q, query_chunk_size))
    denominator, max_score = map(lambda x: rearrange(x, 'b h n c -> n b h c'), (denominator, max_score))

    scale = jnp.sqrt(q.shape[-1])
    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None
    _chunk_bias_fn = partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attn_pdrop, causal, dtype)

    def scan_attention(_, scan):
        q_chunk, numerator_chunk, denominator_chunk, max_score_chunk, q_chunk_idx = scan

        @partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, scan):
            k_chunk, value_chunk, k_chunk_idx = scan
            numerator_chunk, denominator_chunk, prev_max_score_chunk = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bhqk', q_chunk, k_chunk, precision=precision) / scale
            bias_chunk = _chunk_bias_fn(q_chunk_idx_start + q_chunk_idx, k_chunk_idx_start + k_chunk_idx)
            attn_weights = attn_weights + bias_chunk

            max_score_chunk = jnp.maximum(prev_max_score_chunk, jnp.max(attn_weights, axis=-1))
            max_score_chunk = lax.stop_gradient(max_score_chunk)
            exp_weights = jnp.exp(attn_weights - max_score_chunk[..., None])
            exp_values = jnp.einsum('bhqk,bkhd->bqhd', exp_weights, value_chunk, precision=precision)
            correction = rearrange(jnp.exp(prev_max_score_chunk - max_score_chunk), 'b h q -> b q h')[..., None]
            numerator_chunk = numerator_chunk * correction + exp_values
            denominator_chunk = denominator_chunk * jnp.exp(prev_max_score_chunk - max_score_chunk) + exp_weights.sum(
                axis=-1)
            return (numerator_chunk, denominator_chunk, max_score_chunk), None

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, k_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = q_chunk_idx_start + q_chunk_idx < k_chunk_idx_start + k_chunk_idx
            return jax.lax.cond(
                skip_block,
                lambda carry, args: (carry, None),
                scan_kv_block,
                carry,
                args
            )

        (numerator_chunk, denominator_chunk, max_score_chunk), _ = lax.scan(
            skip_upper_half, init=(numerator_chunk, denominator_chunk, max_score_chunk),
            xs=(k, v, jnp.arange(0, num_kv))
        )
        # output_chunk = (numerator_chunk / denominator_chunk).astype(dtype)
        output_chunk = numerator_chunk / rearrange(denominator_chunk, 'b h q -> b q h')[..., None].astype(dtype)
        return (), (output_chunk, numerator_chunk, denominator_chunk, max_score_chunk)

    _, (_, numerator, denominator, max_score) = lax.scan(scan_attention, init=(), xs=(
        q, numerator, denominator, max_score, jnp.arange(0, num_q)))

    numerator = jnp.moveaxis(numerator, 1, 0)
    numerator = numerator.reshape((batch, q_len, num_heads, dim_per_head))
    denominator, max_score = map(lambda x: rearrange(x, 'n b h c -> b h n c'), (denominator, max_score))
    denominator = denominator.reshape((batch, num_heads, q_len))
    max_score = max_score.reshape((batch, num_heads, q_len))

    return numerator, denominator, max_score


def _blockwise_attention_bwd(q, k, v, g, carry, q_chunk_idx_start, k_chunk_idx_start, bias, causal, query_chunk_size,
                             key_chunk_size, deterministic, dropout_rng, attn_pdrop, dtype, policy, precision,
                             prevent_cse):
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    batch, kv_len, num_heads, dim_per_head = v.shape
    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    dq, dk, dv, output, denominator, max_score = carry

    g = g.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    dq = dq.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    dk = dk.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    dv = dv.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    output = output.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    g, dq, dk, dv, output = map(lambda x: jnp.moveaxis(x, 1, 0), (g, dq, dk, dv, output))

    denominator = denominator.reshape((batch, num_heads, num_q, query_chunk_size))
    max_score = max_score.reshape((batch, num_heads, num_q, query_chunk_size))
    denominator, max_score = map(lambda x: rearrange(x, 'b h n c -> n b h c'), (denominator, max_score))

    q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    k = k.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    v = v.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    q, k, v = map(lambda x: jnp.moveaxis(x, 1, 0), (q, k, v))

    scale = jnp.sqrt(q.shape[-1])
    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None
    _chunk_bias_fn = partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attn_pdrop, causal, dtype)

    def scan_attention(carry, scan):
        dk, dv = carry
        q_chunk, dq_chunk, g_chunk, output_chunk, denominator_chunk, max_score_chunk, q_chunk_idx = scan
        dl_part = jnp.einsum("bqhd,bqhd->bhq", g_chunk, output_chunk)[..., None]

        @partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, scan):
            k_chunk, value_chunk, k_chunk_idx = scan
            dq_chunk = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bhqk', q_chunk, k_chunk, precision=precision) / scale
            bias_chunk = _chunk_bias_fn(q_chunk_idx_start + q_chunk_idx, k_chunk_idx_start + k_chunk_idx)
            attn_weights = attn_weights + bias_chunk
            exp_weights = jnp.exp(attn_weights - max_score_chunk[..., None]) / denominator_chunk[..., None]

            ds = jnp.einsum("bqhd,bkhd->bhqk", g_chunk, value_chunk)
            dl = (ds - dl_part) * exp_weights
            dq_chunk = dq_chunk + jnp.einsum("bhqk,bkhd->bqhd", dl, k_chunk) / scale
            dk_chunk = jnp.einsum("bqhd,bhqk->bkhd", q_chunk, dl) / scale
            dv_chunk = jnp.einsum("bhqk,bqhd->bkhd", exp_weights, g_chunk)
            return dq_chunk, (dk_chunk, dv_chunk)

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, k_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = q_chunk_idx_start + q_chunk_idx < k_chunk_idx_start + k_chunk_idx
            return lax.cond(
                skip_block,
                lambda carry, args: (
                    carry, (
                        jnp.zeros((batch, key_chunk_size, num_heads, dim_per_head), dtype=dk.dtype),
                        jnp.zeros((batch, key_chunk_size, num_heads, dim_per_head), dtype=dk.dtype),
                    )
                ),
                scan_kv_block,
                carry,
                args
            )

        dq_chunk, (dk_part, dv_part) = lax.scan(
            skip_upper_half, init=dq_chunk, xs=(k, v, jnp.arange(0, num_kv))
        )
        return (dk + dk_part, dv + dv_part), dq_chunk

    (dk, dv), dq = lax.scan(scan_attention, init=(dk, dv),
                            xs=(q, dq, g, output, denominator, max_score, jnp.arange(0, num_q)))

    dq, dk, dv = map(lambda x: jnp.moveaxis(x, 1, 0), (dq, dk, dv))
    dq = dq.reshape((batch, q_len, num_heads, dim_per_head))
    dk = dk.reshape((batch, kv_len, num_heads, dim_per_head))
    dv = dv.reshape((batch, kv_len, num_heads, dim_per_head))

    return dq, dk, dv


'''
Computing ffn blockwise without materializing the large hidden tensor, training 4x longer sequences than the memory-efficient transformer.
Blockwise parallel transformer https://arxiv.org/abs/2305.19370 Liu et al. 2023
'''


def blockwise_ffn(remat_ffn, inputs, chunk_size, deterministic):
    inputs = rearrange(inputs, 'b (c n) d -> b c n d', c=chunk_size)

    def scan_ffn(remat_ffn, carry, hidden_states):
        outputs = remat_ffn(hidden_states, deterministic=deterministic)
        return carry, outputs

    scan_axis = inputs.ndim - 2
    _, output = nn.scan(
        scan_ffn,
        variable_broadcast="params",
        split_rngs={"params": False, "dropout": True},
        in_axes=scan_axis,
        out_axes=scan_axis,
    )(remat_ffn, None, inputs)
    output = rearrange(output, 'b c n d -> b (c n) d')
    return output


def blockwise_attn(query, key, value, bias, deterministic,
                   dropout_rng, attn_pdrop, causal, query_chunk_size,
                   key_chunk_size, dtype, policy, precision, float32_logits,
                   prevent_cse):
    query = query / jnp.sqrt(query.shape[-1]).astype(dtype)
    if float32_logits:
        query = query.astype(jnp.float32)
        key = key.astype(jnp.float32)

    batch, q_len, num_heads, dim_per_head = query.shape
    batch, kv_len, num_heads, dim_per_head = key.shape
    batch, kv_len, num_heads, dim_per_head = value.shape

    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    query = query.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    key = key.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    value = value.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))

    query = jnp.moveaxis(query, 1, 0)
    key = jnp.moveaxis(key, 1, 0)
    value = jnp.moveaxis(value, 1, 0)

    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None

    _chunk_bias_fn = partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attn_pdrop, causal, dtype)

    def scan_attention(carry, args):
        del carry
        query_chunk, query_chunk_idx = args

        @partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args
            (numerator, denominator, prev_max_score) = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bqhk', query_chunk, key_chunk, precision=precision)
            bias_chunk = _chunk_bias_fn(query_chunk_idx, key_chunk_idx)
            bias_chunk = jnp.moveaxis(bias_chunk, 1, 2)
            attn_weights = attn_weights + bias_chunk

            max_score = jnp.max(attn_weights, axis=-1, keepdims=True)
            max_score = jnp.maximum(prev_max_score, max_score)
            max_score = lax.stop_gradient(max_score)
            exp_weights = jnp.exp(attn_weights - max_score)
            exp_values = jnp.einsum(
                'bqhv,bvhd->bqhd', exp_weights, value_chunk, precision=precision
            )
            correction = jnp.exp(prev_max_score - max_score)
            numerator = numerator * correction + exp_values
            denominator = denominator * correction + exp_weights.sum(axis=-1, keepdims=True)
            return (numerator, denominator, max_score), None

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = query_chunk_idx < key_chunk_idx
            return lax.cond(
                skip_block,
                lambda carry, args: (carry, None),
                scan_kv_block,
                carry,
                args,
            )

        init_carry = (
            jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
            jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
            (-jnp.inf) * jnp.ones((batch, query_chunk_size, num_heads, 1), dtype=query.dtype),
        )
        (numerator, denominator, max_score), _ = lax.scan(
            skip_upper_half, init_carry, xs=(key, value, jnp.arange(0, num_kv))
        )
        output = (numerator / denominator).astype(dtype)
        return (), output

    _, output = lax.scan(scan_attention, (), xs=(query, jnp.arange(0, num_q)))
    output = rearrange(output, 'n b c h d -> b (n c) h d')
    return output


def _chunk_attention_bias(query_chunk_size, key_chunk_size,
                          bias, deterministic, attn_dropout, attn_pdrop, causal,
                          dtype, query_chunk_idx, key_chunk_idx):
    query_offset = query_chunk_idx * query_chunk_size
    key_offset = key_chunk_idx * key_chunk_size
    chunk_bias = jnp.zeros((1, 1, 1, 1), dtype=dtype)
    if bias is not None:
        chunk_bias = lax.dynamic_slice(
            bias,
            start_indices=(0, 0, query_offset, key_offset),
            slice_sizes=(*bias.shape[:2], min(bias.shape[-2], query_chunk_size), min(bias.shape[-1], key_chunk_size)),
        )

    if causal:
        query_idx = lax.broadcasted_iota(dtype=jnp.int32, shape=(query_chunk_size, 1), dimension=0)
        key_idx = lax.broadcasted_iota(dtype=jnp.int32, shape=(1, key_chunk_size), dimension=1)
        offset = query_offset - key_offset
        query_idx += offset
        causal_mask_value = (query_idx < key_idx) * jnp.finfo(dtype).min
        chunk_bias += causal_mask_value.reshape(1, 1, *causal_mask_value.shape)

    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_slice = lax.dynamic_slice(
            attn_dropout,
            start_indices=(0, 0, query_offset, key_offset),
            slice_sizes=(
                *attn_dropout.shape[:2],
                min(attn_dropout.shape[-2], query_chunk_size),
                min(attn_dropout.shape[-1], key_chunk_size),
            ),
        )
        chunk_bias += attn_dropout_slice * jnp.finfo(dtype).min
    return chunk_bias.astype(dtype)


def ring_flash_dummy(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    from jax.experimental.pallas.ops.tpu.flash_attention import flash_attention, BlockSizes, mha_reference
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    q, k, v = map(lambda x: rearrange(x, 'b q h d -> b h q d'), [q, k, v])
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    block_sizes = BlockSizes(
        block_q=query_chunk_size,
        block_k_major=key_chunk_size,
        block_k=key_chunk_size,
        block_b=1,
        block_q_major_dkv=query_chunk_size,
        block_k_major_dkv=key_chunk_size,
        block_k_dkv=key_chunk_size,
        block_q_dkv=query_chunk_size,
        block_k_major_dq=key_chunk_size,
        block_k_dq=key_chunk_size,
        block_q_dq=query_chunk_size,
    )

    o = flash_attention(
        q, k, v,
        causal=blockwise_kwargs["causal"],
        block_sizes=block_sizes,
        sm_scale=q.shape[-1] ** -0.5,
    )
    output = rearrange(o.astype(v.dtype), 'b h q d -> b q h d')
    return output


def ring_standard_dummy(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    from jax.experimental.pallas.ops.tpu.flash_attention import flash_attention, BlockSizes, mha_reference
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    q, k, v = map(lambda x: rearrange(x, 'b q h d -> b h q d'), [q, k, v])
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    o = mha_reference(
        q, k, v,
        ab=None,
        causal=blockwise_kwargs["causal"],
        sm_scale=q.shape[-1] ** -0.5,
    )
    output = rearrange(o.astype(v.dtype), 'b h q d -> b q h d')
    return output


def _ring_flash_attention_fwd_tpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    q, k, v = map(lambda x: rearrange(x, 'b q h d -> b h q d'), [q, k, v])
    batch, num_heads, q_len, dim_per_head = q.shape
    batch, num_heads, kv_len, dim_per_head = k.shape
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    o = jnp.zeros((batch, num_heads, q_len, dim_per_head)).astype(q.dtype)
    l = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
    m = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)

    axis_size = lax.psum(1, axis_name)
    block_size = q_len  # assumes this function is pre-sharded inside shard_map
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    block_sizes = BlockSizes(
        block_q=query_chunk_size,
        block_k_major=key_chunk_size,
        block_k=key_chunk_size,
        block_b=1,
        block_q_major_dkv=query_chunk_size,
        block_k_major_dkv=key_chunk_size,
        block_k_dkv=key_chunk_size,
        block_q_dkv=query_chunk_size,
        block_k_major_dq=key_chunk_size,
        block_k_dq=key_chunk_size,
        block_q_dq=query_chunk_size,
    )

    scale = q.shape[-1] ** -0.5

    def scan_kv_block(carry, idx):
        o, l, m, k, v = carry
        attn_bias_slice = None
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        o, l, m = _flash_attention_fwd(
            q, k, v,
            carry=(o, l, m),
            q_chunk_idx_start=q_chunk_idx_start,
            k_chunk_idx_start=k_chunk_idx_start,
            ab=attn_bias_slice,
            segment_ids=None,
            save_residuals=False,
            causal=blockwise_kwargs["causal"],
            sm_scale=scale,
            block_sizes=block_sizes,
            debug=False
        )
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]),
                   (k, v))
        return (o, l, m, k, v), None

    (o, l, m, _, _), _ = lax.scan(scan_kv_block,
                                  init=(o, l, m, k, v), xs=jnp.arange(0, axis_size))
    output = rearrange(o.astype(v.dtype), 'b h q d -> b q h d')
    return output, (o, q, k, v, attn_bias, l, m)


def _ring_flash_attention_bwd_tpu(axis_name, float32_logits, blockwise_kwargs, res, g):
    del float32_logits
    o, q, k, v, attn_bias, l, m = res
    batch, num_heads, kv_len, dim_per_head = k.shape
    axis_size = lax.psum(1, axis_name)
    dq = jnp.zeros_like(q, dtype=jnp.float32)
    dk = jnp.zeros_like(k, dtype=jnp.float32)
    dv = jnp.zeros_like(v, dtype=jnp.float32)
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]
    block_size = q.shape[2]  # assumes this function is pre-sharded inside shard_map
    scale = q.shape[-1] ** -0.5

    g = rearrange(g, 'b q h d -> b h q d')

    block_sizes = BlockSizes(
        block_q=query_chunk_size,
        block_k_major=key_chunk_size,
        block_k=key_chunk_size,
        block_b=1,
        block_q_major_dkv=query_chunk_size,
        block_k_major_dkv=key_chunk_size,
        block_k_dkv=key_chunk_size,
        block_q_dkv=query_chunk_size,
        block_k_major_dq=key_chunk_size,
        block_k_dq=key_chunk_size,
        block_q_dq=query_chunk_size,
    )

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        # attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1
        # )
        attn_bias_slice = None  # TODO
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        dq_i, dk_i, dv_i, = _flash_attention_bwd(
            save_residuals=False,
            causal=blockwise_kwargs["causal"],
            sm_scale=scale,
            block_sizes=block_sizes,
            debug=False,
            q_chunk_idx_start=q_chunk_idx_start,
            k_chunk_idx_start=k_chunk_idx_start,
            residuals=(q, k, v, attn_bias_slice, None, o, l, m),
            do=g
        )
        dq += dq_i
        dk += dk_i
        dv += dv_i
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
    dq, dk, dv = map(lambda x: rearrange(x, 'b h q d -> b q h d'), (dq, dk, dv))
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6])
def ring_flash_attention_tpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    y, _ = _ring_flash_attention_fwd_tpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs)
    return y


ring_flash_attention_tpu.defvjp(_ring_flash_attention_fwd_tpu, _ring_flash_attention_bwd_tpu)


def ring_flash_dummy_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    output = mha(q, k, v, None, sm_scale=q.shape[-1] ** -0.5)
    return output


def _ring_flash_attention_fwd_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    o = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(jnp.bfloat16)
    l = jnp.zeros((batch, num_heads, q_len)).astype(jnp.float32)
    m = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(jnp.float32)

    axis_size = lax.psum(1, axis_name)
    block_size = q_len  # assumes this function is pre-sharded inside shard_map
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    scale = q.shape[-1] ** -0.5

    def scan_kv_block(carry, idx):
        o, l, m, k, v = carry
        # attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1
        # )
        attn_bias_slice = None  # TODO
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        o, l, m = _mha_forward(
            q, k, v,
            carry=(o.astype(jnp.float32), l, m),
            q_chunk_idx_start=q_chunk_idx_start,  # TODO handle this
            k_chunk_idx_start=k_chunk_idx_start,
            segment_ids=None,
            sm_scale=scale,
            causal=blockwise_kwargs["causal"],
            block_q=query_chunk_size,
            block_k=key_chunk_size,
            backward_pass_impl='triton',  # unused
            num_warps=None,
            num_stages=2,  # unused
            grid=None,
            interpret=False,
            debug=False
        )
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]),
                   (k, v))
        return (o, l, m, k, v), None

    (o, l, m, _, _), _ = lax.scan(scan_kv_block,
                                  init=(o, l, m, k, v), xs=jnp.arange(0, axis_size))
    output = o.astype(v.dtype)
    return output, (o, q, k, v, attn_bias, l, m)


from jax.experimental.pallas.ops.attention import _mha_forward as _mha_forward_d, _mha_backward as _mha_backward_d, mha


def _ring_flash_attention_bwd_gpu(axis_name, float32_logits, blockwise_kwargs, res, g):
    del float32_logits
    o, q, k, v, attn_bias, l, m = res
    batch, kv_len, num_heads, dim_per_head = k.shape
    axis_size = lax.psum(1, axis_name)
    dq = jnp.zeros_like(q, dtype=q.dtype)
    dk = jnp.zeros_like(k, dtype=k.dtype)
    dv = jnp.zeros_like(v, dtype=v.dtype)
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]
    block_size = q.shape[1]  # assumes this function is pre-sharded inside shard_map
    scale = q.shape[-1] ** -0.5

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        # attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1
        # )
        attn_bias_slice = None  # TODO
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        dq_i, dk_i, dv_i = _mha_backward(
            sm_scale=scale,
            causal=blockwise_kwargs["causal"],
            block_q=query_chunk_size,
            block_k=key_chunk_size,
            backward_pass_impl='triton',
            num_warps=None,
            num_stages=2,
            grid=None,
            interpret=False,
            debug=False,
            q_chunk_idx_start=q_chunk_idx_start,
            k_chunk_idx_start=k_chunk_idx_start,
            res=(q, k, v, None, o, l, m),
            do=g,
        )
        dq += dq_i
        dk += dk_i
        dv += dv_i
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6])
def ring_flash_attention_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    y, _ = _ring_flash_attention_fwd_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs)
    return y


ring_flash_attention_gpu.defvjp(_ring_flash_attention_fwd_gpu, _ring_flash_attention_bwd_gpu)

if __name__ == '__main__':
    def reference_attn(query, key, value, causal, dtype):
        query = query / jnp.sqrt(query.shape[-1]).astype(dtype)
        logits = jnp.einsum("bqhc,bkhc->bhqk", query, key)
        if causal:
            mask_value = jnp.finfo(logits.dtype).min
            _, q_seq_len, _, _ = query.shape
            _, kv_seq_len, _, _ = key.shape
            mask_shape = (q_seq_len, kv_seq_len)
            row_ids = lax.broadcasted_iota(jnp.int32, mask_shape, 0)
            col_ids = lax.broadcasted_iota(jnp.int32, mask_shape, 1)
            causal_mask = (row_ids < col_ids)[None, None, :, :]
            logits = logits + jnp.where(causal_mask, mask_value, 0.0)
        weights = jax.nn.softmax(logits, axis=-1)
        out = jnp.einsum("bhqk,bkhc->bqhc", weights, value)
        return out


    # random inputs
    shape = (1, 32, 8, 64)
    query = jax.random.normal(jax.random.PRNGKey(0), shape)
    key = jax.random.normal(jax.random.PRNGKey(1), shape)
    value = jax.random.normal(jax.random.PRNGKey(2), shape)

    causal = True
    chunk_size = 4
    policy = jax.checkpoint_policies.nothing_saveable()

    blockwise = blockwise_attn(query, key, value, None, False, None, 0.0, causal, chunk_size, chunk_size, jnp.float32,
                               policy, 'float32', True, False)
    reference = reference_attn(query, key, value, causal, 'float32')

    print('max diff sum:', jnp.abs(reference - blockwise).sum())
    print('max diff ele:', jnp.abs(reference - blockwise).max())



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized dot_general."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import copy
import functools
from typing import Callable, Optional, Union
from . import config
import flax.struct
import jax
from jax import lax
import jax.numpy as jnp
import numpy as onp


@flax.struct.dataclass
class Context:
    key: Optional[jax.Array]
    train_step: Optional[int]


def _context_split(context: Context) -> tuple[Context, Context]:
    def mk_ctx(key):
        return Context(key=key, train_step=context.train_step)

    if context.key is not None:
        key1, key2 = jax.random.split(context.key)
        return mk_ctx(key1), mk_ctx(key2)
    return mk_ctx(None), mk_ctx(None)


def _scale_quant(x, *, cfg, ca, context):
    """The core quantifying function."""
    msg = (
        'use_fake_quant mode is used in tests and it is exactly equal when'
        ' po2_scale == True; Did you forget to set it?'
    )
    assert (not cfg.use_fake_quant) or cfg.po2_scale, msg

    if isinstance(cfg.numerics, config.NoNumerics):
        return x, None, None
    shared_axes = cfg.calib_shared_axes or ca
    bound = cfg.calibration.get_bound(x, shared_axes)
    abs_max_mapped_to = cfg.numerics.abs_val_mapped_to()
    scale = abs_max_mapped_to / bound

    if cfg.po2_scale:
        scale = 2 ** jnp.floor(jnp.log2(scale))
    if cfg.scale_stop_grad:
        scale = lax.stop_gradient(scale)

    x_s = _maybe_mul(x, scale)

    quant = jax.custom_vjp(cfg.numerics.fwd)
    quant.defvjp(cfg.numerics.vjp_fwd, cfg.numerics.vjp_bwd)
    quant = functools.partial(quant, context=context)

    x_q, quant_grad = jax.vjp(quant, x_s)
    inv_scale = _maybe_inv(scale)

    return x_q, inv_scale, quant_grad


def make_fake_quant(cfg: config.Tensor, ca=None):
    def fake_quant(x, context):
        x_q, inv_scale, _ = _scale_quant(x, cfg=cfg, ca=ca, context=context)
        return _maybe_mul(x_q, inv_scale)

    return fake_quant


@flax.struct.dataclass
# It is used only when use_fwd_quant = True
class QTensor:
    qvalue: jnp.ndarray
    qvalue_scale_t: jnp.ndarray


@flax.struct.dataclass
class MultiTensor:
    x: jnp.ndarray
    qx: Optional[QTensor]


@flax.struct.dataclass
class TensorRes:
    """All the things we pass from the forward pass to the backward pass."""
    mt: MultiTensor
    quant_grad: Union[Callable[[jnp.ndarray], tuple[jnp.ndarray]], None]


@flax.struct.dataclass
class DotGeneralRes:
    context_bwd: Context
    lhs: TensorRes
    rhs: TensorRes


def _scale_trans(x, ca, ba):
    """Transposes x to output dimension order."""
    ca = list(ca)
    ba = list(ba)
    for i in ca:
        assert x.shape[i] == 1
    ra = list(i for i in range(len(x.shape)) if i not in ba + ca)
    x = jnp.transpose(x, ba + ra + ca)
    # x = jnp.squeeze(x, axis=range(len(ba+ra): len(x.shape))
    shape_ba = x.shape[: len(ba)]
    shape_ra = x.shape[len(ba): len(x.shape) - len(ca)]
    # Will need to add additional axes (size 1) for the other shape_ra
    x = x.reshape(shape_ba + shape_ra)
    return x


def _lhs_scale_transpose(lhs_scale, dimension_numbers, lhs_shape, rhs_shape):
    """Transposes lhs_scale to output dimension order."""
    if lhs_scale is None:
        return None

    (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
    qlhs_scale_t = _scale_trans(lhs_scale, lhs_ca, lhs_ba)
    # inserting dummy axes for rhs_ra
    assert len(qlhs_scale_t.shape) == len(lhs_shape) - len(lhs_ca)
    start = len(qlhs_scale_t.shape)
    end = len(rhs_shape) - len(rhs_ca) - len(rhs_ba) + start
    lhs_dummy_axes = range(start, end)
    qlhs_scale_t = jnp.expand_dims(qlhs_scale_t, axis=lhs_dummy_axes)
    return qlhs_scale_t


def _rhs_scale_transpose(rhs_scale, dimension_numbers, lhs_shape, rhs_shape):
    if rhs_scale is None:
        return None
    del rhs_shape
    (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
    qrhs_scale_t = _scale_trans(rhs_scale, rhs_ca, rhs_ba)
    start = len(rhs_ba)
    end = len(lhs_shape) - len(lhs_ca) - len(lhs_ba) + start
    rhs_dummy_axes = range(start, end)
    qrhs_scale_t = jnp.expand_dims(qrhs_scale_t, axis=rhs_dummy_axes)
    return qrhs_scale_t


def _maybe_mul(x, scale):
    if scale is None:
        return x
    return x * scale


def _maybe_inv(x):
    if x is None:
        return None
    return 1.0 / x


def _make_dot_general_raw(gcfg: config.DotGeneralRaw):
    """Makes quantized lax.dot_general replacement."""

    msg = 'Custom calib_shared_axes not implemented for local AQT.'
    assert gcfg.lhs.calib_shared_axes is None, msg
    assert gcfg.rhs.calib_shared_axes is None, msg

    def dot_general_raw(
            lhs: jnp.ndarray,
            rhs: Union[jnp.ndarray, MultiTensor],
            dimension_numbers,
            context,
    ):
        """Creates a dot_general function without custom gradient."""
        (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
        # We need to copy because we modify cfg to populate some defaults.
        cfg = copy.deepcopy(gcfg)

        if isinstance(rhs, MultiTensor):
            # We are in gradient code.
            fwd_quantized = rhs.qx is not None
            expect_fwd_quantized = cfg.rhs.use_fwd_quant is not None
            msg = (
                'Misconfiguration: use_fwd_quant=True, but there is no fwd'
                ' quantization (but rhs.qx is None).'
            )
            assert fwd_quantized == expect_fwd_quantized, msg
            if cfg.rhs.use_fwd_quant:
                assert rhs.qx is not None, msg
                lhs = _maybe_mul(lhs, rhs.qx.qvalue_scale_t)
                rhs = rhs.qx.qvalue
            else:
                rhs = rhs.x
        else:
            assert cfg.rhs.use_fwd_quant is None, 'cannot set use_fwd_quant in fwd'

        if cfg.local_q is not None:

            def factor_reshape(x, ca, ba):
                factor = cfg.local_q.contraction_axis_shard_count
                assert factor is not None
                if len(ca) == 0:
                    return x, ca, ba
                shape = list(x.shape)
                ax = ca[0]
                orig_size = shape[ax]
                assert orig_size % factor == 0
                shape[ax] = factor
                shape.insert(ax + 1, orig_size // factor)
                new_ca = [(b + int(b >= ax)) for b in ca]
                assert new_ca[0] == ax + 1
                new_ba = [ax] + [(b + int(b > ax)) for b in ba]
                return x.reshape(shape), new_ca, new_ba

            lhs, lhs_ca, lhs_ba = factor_reshape(lhs, lhs_ca, lhs_ba)
            rhs, rhs_ca, rhs_ba = factor_reshape(rhs, rhs_ca, rhs_ba)

            dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)

        assert isinstance(rhs, jnp.ndarray)

        context, context_bwd = _context_split(context)
        context_lhs, context_rhs = _context_split(context)
        del context

        lhs_q, lhs_inv_scale, lhs_quant_grad = _scale_quant(
            lhs, cfg=cfg.lhs, ca=lhs_ca, context=context_lhs
        )
        lhs_inv_scale_t = _lhs_scale_transpose(
            lhs_inv_scale, dimension_numbers, lhs.shape, rhs.shape
        )
        lhs_qx = (
            None
            if lhs_inv_scale_t is None
            else QTensor(qvalue=lhs_q, qvalue_scale_t=lhs_inv_scale_t)
        )
        lhs_mt = MultiTensor(x=lhs, qx=lhs_qx)
        lhs_res = TensorRes(mt=lhs_mt, quant_grad=lhs_quant_grad)

        rhs_q, rhs_inv_scale, rhs_quant_grad = _scale_quant(
            rhs, cfg=cfg.rhs, ca=rhs_ca, context=context_rhs
        )
        rhs_inv_scale_t = _rhs_scale_transpose(
            rhs_inv_scale, dimension_numbers, lhs.shape, rhs.shape
        )
        rhs_qx = (
            None
            if rhs_inv_scale_t is None
            else QTensor(qvalue=rhs_q, qvalue_scale_t=rhs_inv_scale_t)
        )
        rhs_mt = MultiTensor(x=rhs, qx=rhs_qx)
        rhs_res = TensorRes(mt=rhs_mt, quant_grad=rhs_quant_grad)

        assert lhs.dtype == rhs.dtype

        if cfg.lhs.use_fake_quant:
            msg = "Can't set dg_in_dtype in fake_quant mode."
            assert cfg.dg_in_dtype is None, msg
            lhs_q = _maybe_mul(lhs_q, lhs_inv_scale)
            rhs_q = _maybe_mul(rhs_q, rhs_inv_scale)
        else:
            if cfg.dg_in_dtype is not None:
                lhs_q = lhs_q.astype(cfg.dg_in_dtype)
                rhs_q = rhs_q.astype(cfg.dg_in_dtype)

        if cfg.lhs.preprocess_quant_cls is not None:
            preprocess_quant_lhs = cfg.lhs.preprocess_quant_cls()
            lhs_q = preprocess_quant_lhs(lhs_q)
        if cfg.lhs.preprocess_scale_cls is not None:
            preprocess_scale_lhs = cfg.lhs.preprocess_scale_cls()
            lhs_inv_scale_t = preprocess_scale_lhs(lhs_inv_scale_t)
        if cfg.rhs.preprocess_quant_cls is not None:
            preprocess_quant_rhs = cfg.rhs.preprocess_quant_cls()
            rhs_q = preprocess_quant_rhs(rhs_q)
        if cfg.rhs.preprocess_scale_cls is not None:
            preprocess_scale_rhs = cfg.rhs.preprocess_scale_cls()
            rhs_inv_scale_t = preprocess_scale_rhs(rhs_inv_scale_t)
        out = lax.dot_general(
            lhs_q,
            rhs_q,
            dimension_numbers=dimension_numbers,
            preferred_element_type=cfg.dg_accumulator_dtype,
            precision=lax.Precision.DEFAULT,
        ).astype(lhs.dtype)

        if not cfg.lhs.use_fake_quant:
            out = _maybe_mul(out, lhs_inv_scale_t)
            out = _maybe_mul(out, rhs_inv_scale_t)

        res = DotGeneralRes(
            context_bwd=context_bwd,
            lhs=lhs_res,
            rhs=rhs_res,
        )
        if cfg.local_q is not None:
            assert len(lhs_ca) == len(rhs_ca)
            if len(lhs_ca) > 0:
                out = jnp.sum(out, axis=0)
            # We are not supporting local AQT in fwd pass, so no res needed.
            res = None
        return out, res

    return dot_general_raw


def _dot_general_raw_attach_gradient(
        fwd_dot_general_raw,
        dlhs_dot_general_raw,
        drhs_dot_general_raw,
):
    """Makes quantized lax.dot_general replacement with attached gradients."""

    def make_fwd(return_residual):
        def fwd(
                lhs,
                rhs,
                dimension_numbers,
                context,
        ):
            assert lhs.dtype == rhs.dtype
            ret, res = fwd_dot_general_raw(
                lhs,
                rhs,
                dimension_numbers,
                context,
            )
            ret = ret.astype(lhs.dtype)
            return (ret, res) if return_residual else ret

        return fwd

    def vjp_bwd(
            fwd_dimension_numbers,
            res: DotGeneralRes,
            g,
    ):
        def ranges_like(*xs):
            start = 0
            for x in xs:
                yield tuple(range(start, start + len(x)))
                start += len(x)

        def grad_dot_general(
                y_res: TensorRes,
                quant_grad,
                dot_general,
                y_is_lhs,
                context,
        ):
            y_ndim = y_res.mt.x.ndim

            (x_ca, y_ca), (x_ba, y_ba) = fwd_dimension_numbers
            if y_is_lhs:
                (y_ca, x_ca) = (x_ca, y_ca)
                (y_ba, x_ba) = (x_ba, y_ba)
            g_ndim = g.ndim - y_ndim + len(x_ba) + 2 * len(x_ca)
            x_ra = tuple(i for i in range(g_ndim) if i not in x_ca and i not in x_ba)
            y_ra = tuple(i for i in range(y_ndim) if i not in y_ca and i not in y_ba)
            if y_is_lhs:
                g_ba, g_ca, _ = ranges_like(x_ba, y_ra, x_ra)
            else:
                g_ba, _, g_ca = ranges_like(x_ba, x_ra, y_ra)
            dims = ((g_ca, y_ra), (g_ba, y_ba))

            out, _ = dot_general(g, y_res.mt, dims, context)

            x_ca_sorted_by_y = tuple(onp.take(x_ca, onp.argsort(y_ca)))
            out_axes = tuple(onp.argsort(tuple(x_ba) + x_ra + x_ca_sorted_by_y))
            transposed_out = jax.lax.transpose(out, out_axes)
            if quant_grad is not None:
                transposed_out = quant_grad(transposed_out)[0]
            return transposed_out

        context1, context2 = _context_split(res.context_bwd)
        dlhs = grad_dot_general(
            res.rhs,
            res.lhs.quant_grad,
            dlhs_dot_general_raw,
            False,
            context1,
        )
        drhs = grad_dot_general(
            res.lhs,
            res.rhs.quant_grad,
            drhs_dot_general_raw,
            True,
            context2,
        )
        return (dlhs, drhs, None)

    vjp = jax.custom_vjp(make_fwd(False), nondiff_argnums=(2,))
    vjp.defvjp(make_fwd(True), vjp_bwd)
    return vjp


def make_dot_general(cfg: Optional[config.DotGeneral]):
    """Makes quantized lax.dot_general replacement with attached gradients."""
    if cfg is None:
        def ret_lax_dg(
                lhs,
                rhs,
                dimension_numbers,
                precision=None,
                preferred_element_type=None,
                *,
                context=Context(key=None, train_step=None),
        ):
            del context
            return jax.lax.dot_general(
                lhs, rhs, dimension_numbers, precision, preferred_element_type
            )

        return ret_lax_dg

    dg = _dot_general_raw_attach_gradient(
        fwd_dot_general_raw=_make_dot_general_raw(cfg.fwd),
        dlhs_dot_general_raw=_make_dot_general_raw(cfg.dlhs),
        drhs_dot_general_raw=_make_dot_general_raw(cfg.drhs),
    )

    def ret_dg(
            lhs,
            rhs,
            dimension_numbers,
            precision=None,
            preferred_element_type=None,
            *,
            context=Context(key=None, train_step=None),
    ):
        del preferred_element_type
        assert (
                precision is None
        ), f'Precision {precision} requested together with quantization.'
        assert lhs.dtype == rhs.dtype, (
            'The only reason we need that, is because we need to determine return'
            ' type.'
        )
        out = dg(
            lhs=lhs,
            rhs=rhs,
            dimension_numbers=dimension_numbers,
            context=context,
        )
        return out

    return ret_dg


def make_conv_general_dilated(cfg: config.DotGeneralRaw):
    """Makes quantized lax.make_conv_general_dilated replacement."""

    cfg = copy.deepcopy(cfg)
    if cfg is None:
        cfg = config.DotGeneralRaw.make()

    def my_conv_general_dilated(
            lhs,
            rhs,
            window_strides,
            padding,
            lhs_dilation=None,
            rhs_dilation=None,
            dimension_numbers=None,
            feature_group_count=1,
            batch_group_count=1,
            precision=None,
            preferred_element_type=None,
    ) -> jax.Array:
        msg1 = """
To simplify the code, we currently assume a Flax-particular layout of the data.
This makes sense, because this is the main use-case of this function.
However if there is any other use, we will drop that assumption.
"""
        rank = len(lhs.shape)
        assert len(rhs.shape) == rank
        assert dimension_numbers is not None, msg1
        assert dimension_numbers.lhs_spec[0:2] == (0, rank - 1), msg1
        assert dimension_numbers.rhs_spec[0:2] == (rank - 1, rank - 2), msg1
        assert dimension_numbers.out_spec[0:2] == (0, rank - 1), msg1
        # In Flax, lhs is the inputs, rhs is the kernel.
        # lhs layout is B, spatials..., Ci
        # rhs layout is: spatials..., Ci, Co
        # out layous it: B, spatials..., Co
        #
        # we need to share these axes: lhs[1:] , rhs[:-1]
        # we have a scale/invscale per: lhs[0] / out[0] and rhs[-1] / out[-1]

        # Flax assumptions.
        assert cfg.lhs.calib_shared_axes == list(range(1, rank))
        assert cfg.rhs.calib_shared_axes == list(range(0, rank - 1))

        lhs_q, lhs_inv_scale, _ = _scale_quant(
            lhs, cfg=cfg.lhs, ca=None, context=None
        )
        rhs_q, rhs_inv_scale, _ = _scale_quant(
            rhs, cfg=cfg.rhs, ca=None, context=None
        )

        out = lax.conv_general_dilated(
            lhs=lhs_q,
            rhs=rhs_q,
            window_strides=window_strides,
            padding=padding,
            lhs_dilation=lhs_dilation,
            rhs_dilation=rhs_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=feature_group_count,
            batch_group_count=batch_group_count,
            precision=precision,
            preferred_element_type=preferred_element_type,
        )

        out = _maybe_mul(out, lhs_inv_scale)
        out = _maybe_mul(out, rhs_inv_scale)

        return out

    return my_conv_general_dilated



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing fused attention forward and backward pass."""
from __future__ import annotations

import functools
from typing import Any, Optional

import jax
from jax import lax
from jax.experimental import pallas as pl
import jax.numpy as jnp


DEFAULT_MASK_VALUE = -0.7 * float(jnp.finfo(jnp.dtype("float32")).max)


def mha_forward_kernel(
        q_ref,
        k_ref,
        v_ref,
        acc_ref,
        l_ref,
        m_ref,  # Input arrays
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        segment_ids_ref: jax.Array | None,  # segment_id arrays
        o_ref: Any,  # Output
        *residual_refs: Any,  # Residual outputs
        num_heads: int,
        sm_scale: float,
        causal: bool,
        block_q: int,
        block_d: int,
        block_k: int,
):
    seq_len = q_ref.shape[0]
    start_q = pl.program_id(0)

    # o is the buffer where we accumulate the output on sram.
    # m_i and l_i (see FlashAttention paper) are updated during the k,v loop.
    m_i = pl.load(m_ref, (pl.dslice(start_q * block_q, block_q),))
    l_i = pl.load(l_ref, (pl.dslice(start_q * block_q, block_q),))
    # acc is the buffer where we accumulate the output on sram.
    o = pl.load(acc_ref, (pl.dslice(start_q * block_q, block_q), pl.dslice(None)))

    q_chunk_idx_start = pl.load(q_chunk_idx_start_ref, (0,))
    k_chunk_idx_start = pl.load(k_chunk_idx_start_ref, (0,))

    # Load q: it will stay in L1 throughout. Indices form a matrix because we
    # read, compute, and write all in 2d chunks. 1 element ~= 1 CUDA thread index.
    # q tile has shape [block_q, block_d], block_d == head_dim.
    curr_q_slice = pl.dslice(start_q * block_q, block_q)
    q = pl.load(q_ref, (curr_q_slice, pl.dslice(None)))
    q_segment_ids = (
        None
        if segment_ids_ref is None
        else pl.load(segment_ids_ref, (curr_q_slice,))
    )

    # In FlashAttention algorithm 1 there are 2 loops: slow over tiles of kv (size
    # (Bc == block_k here), and fast over blocks of q (size Br == block_q here).
    # Here we only loop over blocks of kv to process entire seq_len, the loop over
    # blocks of q is carried out by the grid.
    def body(start_k, carry):
        o_prev, m_prev, l_prev = carry
        curr_k_slice = pl.dslice(start_k * block_k, block_k)

        k = pl.load(k_ref, (curr_k_slice, slice(None)))
        kv_segment_ids = (
            None
            if segment_ids_ref is None
            else pl.load(segment_ids_ref, (curr_k_slice,))
        )
        qk = pl.dot(q, k.T)  # [block_q, block_k]
        if sm_scale != 1.:
            qk *= sm_scale  # [block_q, block_k]

        # Avoids Triton crash.
        # if num_heads > 2:
        #   qk = qk.astype(q_ref.dtype)
        #   qk = qk.astype(jnp.float32)

        if causal or segment_ids_ref is not None:
            mask = None
            if segment_ids_ref is not None:
                mask = segment_mask(q_segment_ids, kv_segment_ids)
            if causal:
                span_q = (q_chunk_idx_start + start_q) * block_q + jnp.arange(block_q)
                span_k = (k_chunk_idx_start + start_k) * block_k + jnp.arange(block_k)
                causal_mask = span_q[:, None] >= span_k[None, :]
                mask = (
                    causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
                )
            # Apply mask to qk.
            qk = jnp.where(mask, qk, DEFAULT_MASK_VALUE)

        m_curr = qk.max(axis=-1)
        m_next = jnp.maximum(m_prev, m_curr)
        correction = jnp.exp(m_prev - m_next)
        l_prev_corr = correction * l_prev
        s_curr = jnp.exp(
            qk - m_next[:, None]
        )  # Use m_next instead of m_curr to avoid a correction on l_curr
        l_curr = s_curr.sum(axis=-1)
        l_next = l_prev_corr + l_curr
        l_next_rcp = 1. / l_next
        s_curr = s_curr * l_next_rcp[:, None]
        o_prev_corr = (l_prev_corr * l_next_rcp)[:, None] * o_prev
        v = pl.load(v_ref, (curr_k_slice, pl.dslice(block_d)))
        o_curr = pl.dot(s_curr.astype(v.dtype), v)

        o_next = o_prev_corr + o_curr
        return o_next, m_next, l_next

    if causal:
        # Ceildiv (`pl.cdiv` and `//` do not work due to type of start_q)
        upper_bound = lax.min(
            lax.div(block_q * (q_chunk_idx_start + start_q + 1) + block_k - 1, block_k) - k_chunk_idx_start,
            pl.cdiv(seq_len, block_k))
    else:
        upper_bound = pl.cdiv(seq_len, block_k)  # type: ignore
    o, m_i, l_i = lax.fori_loop(0, upper_bound, body, (o, m_i, l_i))

    if residual_refs:
        l_ref, m_ref = residual_refs
        pl.store(l_ref, (curr_q_slice,), l_i)
        pl.store(m_ref, (curr_q_slice,), m_i)
    # Write output to dram.
    o = o.astype(o_ref.dtype)
    pl.store(o_ref, (curr_q_slice, pl.dslice(None)), o)


def segment_mask(
        q_segment_ids: jax.Array,
        kv_segment_ids: jax.Array,
):
    # [B, T, 1] or [T, 1]
    q_segment_ids = jnp.expand_dims(q_segment_ids, axis=-1)
    # [B, 1, S] or [1, S]
    if kv_segment_ids.ndim == 1:
        kv_segment_ids = jnp.expand_dims(kv_segment_ids, axis=0)
    else:
        kv_segment_ids = jnp.expand_dims(kv_segment_ids, axis=1)
    return jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)


def _mha_forward(
        q,
        k,
        v,
        carry,
        q_chunk_idx_start,
        k_chunk_idx_start,
        segment_ids: jax.Array | None,
        sm_scale: float,
        causal: bool,
        block_q: int,
        block_k: int,
        backward_pass_impl: str,
        num_warps: Optional[int],
        num_stages: int,
        grid: Any,
        interpret: bool,
        debug: bool,
):
    del backward_pass_impl
    batch_size, seq_len, num_heads, head_dim = q.shape
    block_q = min(block_q, seq_len)
    block_k = min(block_k, seq_len)
    # Heuristics.
    grid_ = grid
    if grid_ is None:
        grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

    num_warps_ = num_warps
    if num_warps_ is None:
        num_warps_ = 4 if head_dim <= 64 else 8
    kernel = functools.partial(mha_forward_kernel, num_heads=num_heads,
                               sm_scale=sm_scale, causal=causal, block_q=block_q,
                               block_k=block_k, block_d=head_dim)
    out_shape = [
        jax.ShapeDtypeStruct(shape=q.shape, dtype=q.dtype),  # out
        jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len),  # l
                             dtype=jnp.float32),
        jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len),  # m
                             dtype=jnp.float32)
    ]
    in_specs = [
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # q
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # k
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # v
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # acc
        pl.BlockSpec(
            lambda _, j, k: (j, k, 0), (None, None, seq_len)
        ),  # l
        pl.BlockSpec(
            lambda _, j, k: (j, k, 0), (None, None, seq_len)
        ),  # m
    ]
    in_specs.append(pl.BlockSpec(lambda _, j, k: (0,), (1,)))
    in_specs.append(pl.BlockSpec(lambda _, j, k: (0,), (1,)))
    in_specs.append(
        None  # type: ignore[arg-type]
        if segment_ids is None
        else pl.BlockSpec(lambda _, j, k: (j, 0), (None, seq_len))
    )
    out, l, m = pl.pallas_call(
        kernel,
        grid=grid_,
        in_specs=in_specs,
        out_specs=[
            pl.BlockSpec(
                lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        num_warps=num_warps_,
        num_stages=num_stages,
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="mha_forward",
    )(q, k, v, *carry, q_chunk_idx_start[None], k_chunk_idx_start[None], segment_ids)
    return out, l, m


def _preprocess_backward_kernel(out_ref, dout_ref, l_ref,
                                new_dout_ref, delta_ref, *,
                                block_q: int):
    pid_m = pl.program_id(0)

    off_m = pl.ds(pid_m * block_q, block_q)
    # load
    o = pl.load(out_ref, (off_m, slice(None))).astype(jnp.float32)
    do = pl.load(dout_ref, (off_m, slice(None))).astype(jnp.float32)
    denom = pl.load(l_ref, (off_m,)).astype(jnp.float32)
    # compute
    do = do / denom[:, None]
    delta = jnp.sum(o * do, axis=1)
    # write-back
    pl.store(new_dout_ref, (off_m, slice(None)),
             do.astype(new_dout_ref.dtype))
    pl.store(delta_ref, (off_m,), delta.astype(delta_ref.dtype))


@jax.named_scope("preprocess_backward")
def _preprocess_backward(out, do, l, block_q: int,
                         debug: bool, interpret: bool):
    batch_size, seq_len, num_heads, head_dim = out.shape
    out_shape = [
        jax.ShapeDtypeStruct(do.shape, do.dtype),
        jax.ShapeDtypeStruct(l.shape, l.dtype),
    ]
    do_scaled, delta = pl.pallas_call(
        functools.partial(_preprocess_backward_kernel, block_q=block_q),
        grid=(pl.cdiv(seq_len, block_q), batch_size, num_heads),
        in_specs=[
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        out_specs=[
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        num_warps=4,
        num_stages=3,
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="mha_preprocess_backward")(out, do, l)
    return do_scaled, delta


def mha_backward_kernel(
        # Inputs
        q_ref,
        k_ref,
        v_ref,
        segment_ids_ref: jax.Array | None,
        out_ref,
        do_scaled_ref,
        l_ref,
        m_ref,
        delta_ref,
        _,
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        # Outputs
        dq_ref,
        dk_ref,
        dv_ref,
        *,
        sm_scale: float,
        causal: bool,
        block_q: int,
        block_d: int,
        block_k: int,
):
    del out_ref, l_ref  # Not needed
    seq_len = q_ref.shape[0]
    q_chunk_idx_start = pl.load(q_chunk_idx_start_ref, (0,))
    k_chunk_idx_start = pl.load(k_chunk_idx_start_ref, (0,))

    def outer_loop(start_k, _):

        dv = jnp.zeros([block_k, block_d], dtype=jnp.float32)
        dk = jnp.zeros([block_k, block_d], dtype=jnp.float32)
        k = pl.load(k_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
        v = pl.load(v_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
        span_k = (k_chunk_idx_start + start_k) * block_k + jnp.arange(block_k)
        kv_segment_ids = (
            None
            if segment_ids_ref is None
            else pl.load(segment_ids_ref, (pl.ds(start_k * block_k, block_k),))
        )

        def inner_loop(start_q, carry):
            dv, dk = carry
            q = pl.load(q_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
            qk = pl.dot(q, k.T)
            qk = qk.astype(q_ref.dtype)
            qk = qk.astype(jnp.float32)
            if sm_scale != 1.0:
                qk *= sm_scale

            q_segment_ids = (
                None
                if segment_ids_ref is None
                else pl.load(segment_ids_ref, (pl.ds(start_q * block_q, block_q),))
            )

            if causal or segment_ids_ref is not None:
                mask = None
                if segment_ids_ref is not None:
                    mask = segment_mask(q_segment_ids, kv_segment_ids)

                if causal:
                    span_q = (q_chunk_idx_start + start_q) * block_q + jnp.arange(block_q)
                    causal_mask = span_q[:, None] >= span_k[None, :]
                    mask = (
                        causal_mask
                        if mask is None
                        else jnp.logical_and(mask, causal_mask)
                    )
                qk = jnp.where(mask, qk, DEFAULT_MASK_VALUE)

            m = pl.load(m_ref, (pl.ds(start_q * block_q, block_q),))
            p = jnp.exp(qk - m[:, None])
            do = pl.load(do_scaled_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
            dv = dv + pl.dot(p.astype(do.dtype).T, do)
            di = pl.load(delta_ref, (pl.ds(start_q * block_q, block_q),))
            dp = jnp.zeros((block_q, block_k), dtype=jnp.float32) - di[:, None]
            dp = dp + pl.dot(do, v.T)
            ds = p * dp
            if sm_scale != 1.0:
                ds = ds * sm_scale
            dk = dk + pl.dot(ds.astype(q_ref.dtype).T, q)
            dq = pl.load(dq_ref, (pl.ds(start_q * block_q, block_q),
                                  slice(None)), eviction_policy="evict_last")
            dq = dq + pl.dot(ds.astype(k.dtype), k).astype(dq.dtype)
            pl.store(dq_ref, (pl.ds(start_q * block_q, block_q),
                              slice(None)), dq, eviction_policy="evict_last")
            return dv, dk

        if causal:
            lower_bound = lax.max(0, lax.div((k_chunk_idx_start + start_k) * block_k, block_q) - q_chunk_idx_start)
        else:
            lower_bound = 0
        dv, dk = lax.fori_loop(lower_bound, pl.cdiv(seq_len, block_q), inner_loop,
                               (dv, dk))
        pl.store(dv_ref, (pl.ds(start_k * block_k, block_k),
                          slice(None)), dv.astype(dv_ref.dtype))
        pl.store(dk_ref, (pl.ds(start_k * block_k, block_k),
                          slice(None)), dk.astype(dk_ref.dtype))

    lax.fori_loop(0, pl.cdiv(seq_len, block_k), outer_loop, None)


def _mha_backward(sm_scale: float, causal: bool, block_q: int, block_k: int,
                  backward_pass_impl: str, num_warps: Optional[int],
                  num_stages: int, grid: Any, interpret: bool,
                  debug: bool, q_chunk_idx_start, k_chunk_idx_start, res, do):
    del num_warps, num_stages, grid
    q, k, v, segment_ids, out, l, m = res

    if backward_pass_impl == "xla":
        raise Exception("use backward_pass_impl == triton")
        # return jax.vjp(
        #     functools.partial(mha_reference, sm_scale=sm_scale, causal=causal),
        #     q,
        #     k,
        #     v,
        #     segment_ids,
        # )[1](do)
    elif backward_pass_impl == "triton":
        batch_size, seq_len, num_heads, head_dim = q.shape
        block_q = min(block_q, seq_len)
        block_k = min(block_k, seq_len)
        do_scaled, delta = _preprocess_backward(out, do, l, block_q, debug, interpret)
        # We accumulate into dq so we need to initialize it to zeros.
        dq = jnp.zeros(q.shape, jnp.float32)
        out_shapes = [
            jax.ShapeDtypeStruct(dq.shape, dq.dtype),
            jax.ShapeDtypeStruct(k.shape, k.dtype),
            jax.ShapeDtypeStruct(v.shape, v.dtype),
        ]

        in_specs = [
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
        ]
        if segment_ids is None:
            in_specs.insert(3, None)  # type: ignore[arg-type]
            input_output_aliases = {8: 0}
        else:
            in_specs.insert(3, pl.BlockSpec(lambda j, k: (j, 0), (None, seq_len)))
            input_output_aliases = {9: 0}
        in_specs.append(pl.BlockSpec(lambda j, k: (0,), (1,)))
        in_specs.append(pl.BlockSpec(lambda j, k: (0,), (1,)))
        grid = (batch_size, num_heads)
        # TODO(sharadmv): figure out why num_warps=8 doesn't work!
        num_warps = 8
        dq, dk, dv = pl.pallas_call(
            functools.partial(
                mha_backward_kernel,
                block_q=block_q,
                block_d=head_dim,
                block_k=block_k,
                sm_scale=sm_scale,
                causal=causal,
            ),
            grid=grid,
            out_shape=out_shapes,
            in_specs=in_specs,
            out_specs=[
                pl.BlockSpec(
                    lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
                ),
                pl.BlockSpec(
                    lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
                ),
                pl.BlockSpec(
                    lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
                ),
            ],
            name="mha_backward",
            debug=debug,
            interpret=interpret,
            num_warps=num_warps,
            num_stages=1,
            input_output_aliases=input_output_aliases,
        )(q, k, v, segment_ids, out, do_scaled, l, m, delta, dq, q_chunk_idx_start[None], k_chunk_idx_start[None])
    else:
        raise ValueError(f"Invalid backward pass implementation: {backward_pass_impl}")
    return dq.astype(q.dtype), dk, dv



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flax layer for AQT injection."""

import enum
import functools
from . import q_dot_general
from . import calibration
from . import config
from . import int_numerics
import flax.linen as nn
import jax.numpy as jnp
from typing import Optional, Union


class Freezer(nn.Module, config.Preprocess):
    """Identity function that can freeze its input.

    On default it is an identity function that saves the input in a variable.
    In 'use_frozen=True' mode, ignores the input and returns the frozen value. It
    is usefult to implement 'constant folding' and put quantized weights and
    scales in the checkpoint for serving.
    """

    # If you want use 'params' make sure that there is another mechanism to hide
    # these variables from the optimizer.
    var_collection: str = 'q'

    # If you set it to True, instead of returning the current input
    # will return last input it got.
    use_frozen: bool = False

    @nn.compact
    def __call__(self, inputs):
        # return inputs or the frozen value
        collection = self.var_collection
        frozen = self.variable(collection, 'frozen', jnp.zeros, inputs.shape)
        if not self.use_frozen:
            frozen.value = inputs
        return frozen.value


class QDotGeneral(nn.Module):
    """A layer that can be injected into flax.nn.Dense, etc."""

    cfg: Optional[Union[config.DotGeneral, None]] = None
    prng_name: Optional[Union[str, None]] = 'params'

    @nn.compact
    def __call__(
            self,
            lhs,
            rhs,
            dimension_numbers,
            precision,
            preferred_element_type=None,
    ):
        key = self.make_rng(self.prng_name) if self.prng_name is not None else None
        context = q_dot_general.Context(key=key, train_step=None)
        q_dg = q_dot_general.make_dot_general(self.cfg)
        q_dg = functools.partial(q_dg, context=context)
        return q_dg(
            lhs,
            rhs,
            dimension_numbers,
            precision,
            preferred_element_type=preferred_element_type,
        )


class QEinsum(nn.Module):
    """Quantized Einsum class for model injection."""

    cfg: Optional[Union[config.DotGeneral, None]] = None
    prng_name: Optional[Union[str, None]] = 'params'

    @nn.compact
    def __call__(self, eqn, lhs, rhs):
        key = self.make_rng(self.prng_name) if self.prng_name is not None else None
        context = q_dot_general.Context(key=key, train_step=None)
        q_dg = q_dot_general.make_dot_general(self.cfg)
        q_dg = functools.partial(q_dg, context=context)
        return jnp.einsum(eqn, lhs, rhs, _dot_general=q_dg)


class QuantMode(enum.Enum):
    DYNAMIC = 1
    FREEZE = 2
    SERVE_FROZEN = 3


def mk_freezer(name: str, freeze_collection: str, mode: QuantMode):
    assert mode in QuantMode
    if mode == QuantMode.DYNAMIC:
        return None
    use_frozen = mode == QuantMode.SERVE_FROZEN
    return functools.partial(
        Freezer,
        name=name,
        use_frozen=use_frozen,
        var_collection=freeze_collection,
    )


def set_rhs_quant_mode(
        cfg: config.DotGeneral, mode: QuantMode, collection='q'
):
    cfg.fwd.rhs.preprocess_quant_cls = mk_freezer('rhs', collection, mode)
    cfg.fwd.rhs.preprocess_scale_cls = mk_freezer('rhs_scale', collection, mode)


def set_lhs_quant_mode(
        cfg: config.DotGeneral, mode: QuantMode, collection='q'
):
    cfg.fwd.lhs.preprocess_quant_cls = mk_freezer('lhs', collection, mode)
    cfg.fwd.lhs.preprocess_scale_cls = mk_freezer('lhs_scale', collection, mode)


def config_v4(
        *,
        fwd_bits: Union[int, None],
        dlhs_bits: Union[int, None],
        drhs_bits: Union[int, None],
        # The dummy static bound flag is for performance benchmarking.
        use_dummy_static_bound: bool = False,
        rng_type: str = 'jax.uniform',  # 'custom-1'
        dlhs_local_q: Union[config.LocalQ, None] = None,
        drhs_local_q: Union[config.LocalQ, None] = None,
        fwd_accumulator_dtype: ... = jnp.int32,
        dlhs_accumulator_dtype: ... = jnp.int32,
        drhs_accumulator_dtype: ... = jnp.int32,
        lhs_quant_mode: QuantMode = QuantMode.DYNAMIC,
        rhs_quant_mode: QuantMode = QuantMode.DYNAMIC,
        freeze_collection: str = 'q',
) -> config.DotGeneral:
    """Version 4 of user-visible AQT config."""

    def tensor_config(bits: Union[int, None]) -> config.Tensor:
        assert bits is None or bits >= 2, 'Need at least 2 bits.'
        if bits is None:
            numerics = config.NoNumerics()
        else:
            numerics = int_numerics.IntNumerics(
                bits=bits,
                preserve_zero=True,
                preserve_max_val=False,
                clip=True,
                round=True,
                noise_fn=None,
                clip_gradient=False,  # Can be False when using abs-max scaling.
            )

        return config.Tensor(
            numerics=numerics,
            calib_shared_axes=None,
            scale_stop_grad=True,
            calibration=calibration.AbsMaxCalibration(),
            po2_scale=False,
            use_fake_quant=False,
            # dtype_x=dtype,
            use_fwd_quant=None,
            preprocess_quant_cls=None,
            preprocess_scale_cls=None,
        )

    def dg_raw_config(lhs_bits, rhs_bits, local_q=None) -> config.DotGeneralRaw:
        lhs_cfg = tensor_config(lhs_bits)
        rhs_cfg = tensor_config(rhs_bits)
        if (
                True  # Just to format lines below
                and lhs_bits is not None
                and rhs_bits is not None
                and lhs_bits <= 8
                and rhs_bits <= 8
        ):
            dg_in_dtype = jnp.int8
            dg_accumulator_dtype = jnp.int32
        else:
            # None determines the dtype on the fly in q_dot_general
            dg_in_dtype = None
            dg_accumulator_dtype = None

        return config.DotGeneralRaw(
            lhs=lhs_cfg,
            rhs=rhs_cfg,
            dg_in_dtype=dg_in_dtype,
            dg_accumulator_dtype=dg_accumulator_dtype,
            local_q=local_q,
        )

    cfg = config.DotGeneral(
        fwd=dg_raw_config(fwd_bits, fwd_bits),
        dlhs=dg_raw_config(dlhs_bits, dlhs_bits, local_q=dlhs_local_q),
        drhs=dg_raw_config(drhs_bits, drhs_bits, local_q=drhs_local_q),
    )

    cfg.dlhs.rhs.use_fwd_quant = False
    cfg.drhs.rhs.use_fwd_quant = False

    # Typically we have (but I don't know if it is guraranteed):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    config.set_stochastic_rounding(
        cfg,
        vjp_lhs_stochastic_rounding=True,
        vjp_rhs_stochastic_rounding=False,
        implementation=rng_type,
    )

    if use_dummy_static_bound:
        config.set_static_bound(cfg, 1.0)

    config.set_accumulator_dtype(
        cfg,
        fwd_dtype=fwd_accumulator_dtype,
        dlhs_dtype=dlhs_accumulator_dtype,
        drhs_dtype=drhs_accumulator_dtype,
    )

    assert (
            lhs_quant_mode == QuantMode.DYNAMIC or rhs_quant_mode == QuantMode.DYNAMIC
    ), (
        'It seems unlikely that both sides of the matmul should be frozen.'
        ' E.g. both sides of the matmul be weights. '
    )

    set_rhs_quant_mode(cfg, rhs_quant_mode, freeze_collection)
    set_lhs_quant_mode(cfg, lhs_quant_mode, freeze_collection)

    return cfg



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""

import abc
import dataclasses
from typing import Any, Callable, Optional, Union
from . import calibration
from . import int_numerics
from . import stochastic_rounding
import flax.struct
import jax
import jax.numpy as jnp

DType = Any
Context = Any  # TODO(lew): We could put Context in a separate file.

ClipAndRoundFn = Callable[[jnp.ndarray, Context], jnp.ndarray]


class Preprocess(abc.ABC):

    @abc.abstractmethod
    def __call__(self, inputs):
        pass


class NoNumerics(flax.struct.PyTreeNode):
    # TODO(lew): This is a workaround. We should separate Stochastic Rounding.
    # noise_fn has no effect in NoNumerics.
    noise_fn: Optional[stochastic_rounding.NoiseFn] = None

    """No quantization, use a native type such as bf16."""
    pass


Numerics = Union[NoNumerics, int_numerics.IntNumerics]


@dataclasses.dataclass
class Tensor:
    """Configuration of quantization of one tensor or one side of tensor op."""

    numerics: Numerics
    calib_shared_axes: Optional[list[int]]
    scale_stop_grad: bool
    # noise+clip+round
    # We apply gradient of clip_and_round in bwd pass.
    calibration: calibration.Calibration
    # Round up the calibration to power of 2 (po2).
    po2_scale: bool
    use_fake_quant: bool
    # Controls at what value of input tensor should be used.
    # Setting it to True, but not quantizing fwd pass will assert-fail.
    use_fwd_quant: Optional[bool]
    # Operation applied to the quantized inputs to the dot general
    preprocess_quant_cls: Optional[Callable[[], Preprocess]]
    # lax.dot_general output is multiplied by (int transposed) scales.
    # preprocess_scale will be applied to these scales before the multiplication.
    preprocess_scale_cls: Optional[Callable[[], Preprocess]]

    @classmethod
    def make(cls, *args, **kwargs) -> 'Tensor':
        return tensor_make(*args, **kwargs)


@dataclasses.dataclass
class LocalQ:
    contraction_axis_shard_count: int


@dataclasses.dataclass
class DotGeneralRaw:
    """Configuration of quantization of one dot_general without gradient."""

    lhs: Tensor
    rhs: Tensor
    dg_in_dtype: Optional[DType]
    dg_accumulator_dtype: Optional[DType]
    local_q: Optional[LocalQ]

    @classmethod
    def make(cls, *args, **kwargs) -> 'DotGeneralRaw':
        return dot_general_raw_make(*args, **kwargs)

    @classmethod
    def make_conv_general_dilated(cls, *args, **kwargs) -> 'DotGeneralRaw':
        return conv_general_dilated_make(*args, **kwargs)


@dataclasses.dataclass
class DotGeneral:
    """Configuration of quantization of dot_general and its gradients."""

    fwd: DotGeneralRaw
    dlhs: DotGeneralRaw
    drhs: DotGeneralRaw

    @classmethod
    def make(cls, *args, **kwargs) -> 'DotGeneral':
        return dot_general_make(*args, **kwargs)


################################################################################
# Functions below are auxiliary helpers.


def set_accumulator_dtype(
        cfg: DotGeneral,
        fwd_dtype: Optional[DType],
        dlhs_dtype: Optional[DType],
        drhs_dtype: Optional[DType],
):
    cfg.fwd.dg_accumulator_dtype = fwd_dtype
    cfg.dlhs.dg_accumulator_dtype = dlhs_dtype
    cfg.drhs.dg_accumulator_dtype = drhs_dtype


def set_stochastic_rounding(
        cfg: DotGeneral,
        # Typically we have (but it's a caller's responsibility to check):
        # - vjp_lhs_stochastic_rounding is referring to the gradient and
        # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
        vjp_lhs_stochastic_rounding: bool,
        vjp_rhs_stochastic_rounding: bool,
        implementation: str,
):
    """Configure stochastic rounding implementation."""
    noise_implementations = {
        'jax.uniform': lambda shape, key: jax.random.uniform(key, shape) - 0.5,
        'custom-1': stochastic_rounding.random_centered_uniform,
    }
    msg = f'{implementation} not supported.'
    assert implementation in noise_implementations.keys(), msg
    noise_fn = noise_implementations[implementation]

    if vjp_lhs_stochastic_rounding:
        cfg.dlhs.lhs.numerics = cfg.dlhs.lhs.numerics.replace(noise_fn=noise_fn)
        cfg.drhs.lhs.numerics = cfg.drhs.lhs.numerics.replace(noise_fn=noise_fn)
    else:
        cfg.dlhs.lhs.numerics = cfg.dlhs.lhs.numerics.replace(noise_fn=None)
        cfg.drhs.lhs.numerics = cfg.drhs.lhs.numerics.replace(noise_fn=None)

    if vjp_rhs_stochastic_rounding:
        cfg.dlhs.rhs.numerics = cfg.dlhs.rhs.numerics.replace(noise_fn=noise_fn)
        cfg.drhs.rhs.numerics = cfg.drhs.rhs.numerics.replace(noise_fn=noise_fn)
    else:
        cfg.dlhs.rhs.numerics = cfg.dlhs.rhs.numerics.replace(noise_fn=None)
        cfg.drhs.rhs.numerics = cfg.drhs.rhs.numerics.replace(noise_fn=None)


def set_static_bound(cfg: DotGeneral, bound: float = 1.0):
    cfg.fwd.lhs.calibration = calibration.ConstantCalibration(bound)
    cfg.fwd.rhs.calibration = calibration.ConstantCalibration(bound)
    cfg.drhs.lhs.calibration = calibration.ConstantCalibration(bound)
    cfg.drhs.rhs.calibration = calibration.ConstantCalibration(bound)
    cfg.dlhs.lhs.calibration = calibration.ConstantCalibration(bound)
    cfg.dlhs.rhs.calibration = calibration.ConstantCalibration(bound)


def tensor_make(bits: Optional[int]) -> 'Tensor':
    """Makes config.Tensor."""
    if bits is None:
        numerics = NoNumerics()
    else:
        pz = False if bits == 1 else True
        numerics = int_numerics.IntNumerics(
            bits=bits,
            preserve_zero=pz,
            preserve_max_val=False,
            clip=True,
            round=True,
            noise_fn=None,
            clip_gradient=False,  # This can be disabled when using abs-max scaling.
        )

    return Tensor(
        numerics=numerics,
        calib_shared_axes=None,
        scale_stop_grad=True,
        calibration=calibration.AbsMaxCalibration(),
        po2_scale=False,
        use_fake_quant=False,
        # dtype_x=dtype,
        use_fwd_quant=None,
        preprocess_quant_cls=None,
        preprocess_scale_cls=None,
    )


def dot_general_raw_make(
        lhs_bits=None,
        rhs_bits=None,
        local_q=None,
) -> 'DotGeneralRaw':
    """Create quantization configs for input matrices to a matmul."""
    lhs_cfg = tensor_make(lhs_bits)
    rhs_cfg = tensor_make(rhs_bits)

    # Binary uses 0.5 right now.
    if (
            lhs_bits is not None
            and rhs_bits is not None
            and 2 <= lhs_bits <= 8
            and 2 <= rhs_bits <= 8
    ):
        dg_in_dtype = jnp.int8
        dg_accumulator_dtype = jnp.int32
    else:
        # Use None to determine the dtype on the fly in q_dot_general
        dg_in_dtype = None
        dg_accumulator_dtype = None

    return DotGeneralRaw(
        lhs=lhs_cfg,
        rhs=rhs_cfg,
        dg_in_dtype=dg_in_dtype,
        dg_accumulator_dtype=dg_accumulator_dtype,
        local_q=local_q,
    )


def conv_general_dilated_make(
        spatial_dimensions=2,
        lhs_bits: Optional[int] = None,
        rhs_bits: Optional[int] = None,
) -> 'DotGeneralRaw':
    """Create quantization config conv_general_dilated."""
    config = dot_general_raw_make(lhs_bits, rhs_bits)
    # Hardcoding flax assumptions.
    if config.lhs:
        config.lhs.calib_shared_axes = list(range(1, spatial_dimensions + 2))
    if config.rhs:
        config.rhs.calib_shared_axes = list(range(0, spatial_dimensions + 2 - 1))
    return config


def dot_general_make(
        lhs_bits: Optional[int] = None,
        rhs_bits: Optional[int] = None,
        bwd_bits: Optional[int] = None,
        use_fwd_quant: bool = True,
        dlhs_local_q=None,
        drhs_local_q=None,
) -> 'DotGeneral':
    """Create quantization configs for input matrices to a matmul."""
    fwd = dot_general_raw_make(lhs_bits, rhs_bits)
    dlhs = dot_general_raw_make(bwd_bits, bwd_bits, local_q=dlhs_local_q)
    drhs = dot_general_raw_make(bwd_bits, bwd_bits, local_q=drhs_local_q)
    cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

    # Surprising: lhs quantization determines what drhs can do.
    if lhs_bits is not None:
        # Only rhs is accepting MultiTensor.
        cfg.drhs.rhs.use_fwd_quant = use_fwd_quant
    if rhs_bits is not None:
        cfg.dlhs.rhs.use_fwd_quant = use_fwd_quant
    return cfg


def fully_quantized(
        *,
        fwd_bits: Union[int, None] = 8,
        bwd_bits: Union[int, None] = 8,
        use_fwd_quant: bool = True,
        use_stochastic_rounding: Optional[bool] = True,
        # Typically we have (but it's a caller's responsibility to check):
        # - vjp_lhs_stochastic_rounding is referring to the gradient and
        # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
        vjp_lhs_stochastic_rounding: Optional[bool] = None,
        vjp_rhs_stochastic_rounding: Optional[bool] = None,
        # The dummy static bound flag is temporary, for performance benchmarking.
        use_dummy_static_bound: bool = False,
        dlhs_local_q: Optional[LocalQ] = None,
        drhs_local_q: Optional[LocalQ] = None,
) -> DotGeneral:
    """Fully Quantized Training."""
    cfg = dot_general_make(
        lhs_bits=fwd_bits,
        rhs_bits=fwd_bits,
        bwd_bits=bwd_bits,
        use_fwd_quant=use_fwd_quant,
        dlhs_local_q=dlhs_local_q,
        drhs_local_q=drhs_local_q,
    )

    # Stochastic Rounding
    # These 3 variables are used to ensure we don't mix
    # old and new style of SR configuration.
    old_style_sr_config = use_stochastic_rounding is not None
    new_style_sr_config_lhs = vjp_lhs_stochastic_rounding is not None
    new_style_sr_config_rhs = vjp_rhs_stochastic_rounding is not None
    assert new_style_sr_config_lhs == new_style_sr_config_rhs, (
        'if you use new style SR config (vjp_xhs_stochastic_rounding), do pass'
        ' both lhs and rhs explicitely.'
    )
    assert new_style_sr_config_lhs != old_style_sr_config

    true = True  # A crude way to get around g-explicit-bool-comparison warning

    assert not (vjp_lhs_stochastic_rounding and vjp_rhs_stochastic_rounding), (
        'This config is buggy when you set both to True. Contact lew@ or use'
        ' config_v3'
    )

    # By default use jax.uniform for stochastic rounding
    if use_stochastic_rounding == true:
        set_stochastic_rounding(cfg, True, True, 'jax.uniform')

    if vjp_lhs_stochastic_rounding == true:
        set_stochastic_rounding(cfg, True, False, 'jax.uniform')

    if vjp_rhs_stochastic_rounding == true:
        set_stochastic_rounding(cfg, False, True, 'jax.uniform')

    if use_dummy_static_bound:
        set_static_bound(cfg, 1.0)

    return cfg


def config_v3(
        *,
        fwd_bits: Union[int, None],
        dlhs_bits: Union[int, None],
        drhs_bits: Union[int, None],
        use_dummy_static_bound: bool = False,
        rng_type: str = 'jax.uniform',  # 'custom-1'
        dlhs_local_q: Optional[LocalQ] = None,
        drhs_local_q: Optional[LocalQ] = None,
        fwd_accumulator_dtype: ... = jnp.int32,
        dlhs_accumulator_dtype: ... = jnp.int32,
        drhs_accumulator_dtype: ... = jnp.int32,
) -> DotGeneral:
    """Fully Quantized Training."""
    fwd = dot_general_raw_make(fwd_bits, fwd_bits)
    dlhs = dot_general_raw_make(dlhs_bits, dlhs_bits, local_q=dlhs_local_q)
    drhs = dot_general_raw_make(drhs_bits, drhs_bits, local_q=drhs_local_q)
    cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

    cfg.dlhs.rhs.use_fwd_quant = False
    cfg.drhs.rhs.use_fwd_quant = False

    set_stochastic_rounding(
        cfg,
        vjp_lhs_stochastic_rounding=True,
        vjp_rhs_stochastic_rounding=False,
        implementation=rng_type,
    )

    if use_dummy_static_bound:
        set_static_bound(cfg, 1.0)

    set_accumulator_dtype(
        cfg,
        fwd_dtype=fwd_accumulator_dtype,
        dlhs_dtype=dlhs_accumulator_dtype,
        drhs_dtype=drhs_accumulator_dtype,
    )
    return cfg



import importlib.util
import jax
import flax


def is_torch_available():
    return True if importlib.util.find_spec('torch') is not None else False


def count_num_params(_p):
    return sum(i.size for i in jax.tree_util.tree_flatten(flax.core.unfreeze(_p))[0])


def count_params(_p):
    print('\033[1;31mModel Contain : ', count_num_params(_p) / 1e9, ' Billion Parameters')


class JaxRNG(object):
    @classmethod
    def from_seed(cls, seed):
        return cls(jax.random.PRNGKey(seed))

    def __init__(self, rng):
        self.rng = rng

    def __call__(self, keys=None):
        if keys is None:
            self.rng, split_rng = jax.random.split(self.rng)
            return split_rng
        elif isinstance(keys, int):
            split_rngs = jax.random.split(self.rng, num=keys + 1)
            self.rng = split_rngs[0]
            return tuple(split_rngs[1:])
        else:
            split_rngs = jax.random.split(self.rng, num=len(keys) + 1)
            self.rng = split_rngs[0]
            return {key: val for key, val in zip(keys, split_rngs[1:])}


def init_rng(seed):
    global jax_utils_rng
    jax_utils_rng = JaxRNG.from_seed(seed)


def next_rng(*args, **kwargs):
    global jax_utils_rng
    return jax_utils_rng(*args, **kwargs)


class GenerateRNG:
    def __init__(self, seed: int = 0):
        self.seed = seed
        self.rng = jax.random.PRNGKey(seed)

    def __next__(self):
        while True:
            self.rng, ke = jax.random.split(self.rng, 2)
            return ke
